# Copyright 2020 Cognicept Systems
# Author: Jakub Tomasek (jakub@cognicept.systems)
# --> AgentLifeCycle handles life cycle of Cognicept Agents

import time
import docker
import boto3
import getpass
import base64
import json
import os
import sys
import subprocess
import dateutil
from datetime import datetime
from ping3 import ping
import re
from docker.errors import DockerException
import glob
from multiprocessing import Process, Queue
import requests
import botocore
import shutil
import pkg_resources
from tabulate import tabulate
from cogniceptshell.common import bcolors
from cogniceptshell.common import generate_progress_bar
from cogniceptshell.common import permission_safe_docker_call
from subprocess import DEVNULL
from docker.types import LogConfig
from compose.cli.main import TopLevelCommand, project_from_options

class AgentLifeCycle:
    """
    A class to manage agent and Cognicept's docker container lifecycle
    ...

    Parameters
    ----------
    None

    Methods
    -------
    configure_containers(cfg):
        Loads agent configuration from `COG_AGENT_CONTAINERS` and `COG_AGENT_IMAGES`
    get_status(args):
        Prints status of docker containers listed in `COG_AGENT_CONTAINERS`.
    get_last_event(args):
        Prints last log in `~/.cognicept/agent/logs/`.
    restart(args):
        Stops and starts the containers listed in `COG_AGENT_CONTAINERS`.
    start(args):
        Starts the containers listed in `COG_AGENT_CONTAINERS`. If `args` has parameter `list`, starts only containers in the list.
    stop(args):
        Stops the containers listed in `COG_AGENT_CONTAINERS`. If `args` has parameter `list`, stops only containers in the list.
    update(args):
        Pulls docker images listed in `COG_AGENT_IMAGES`.
    run_orbitty(args):
        Starts Orbitty.
    """

    # default configuration of containers and images
    _docker_container_names = ["cgs_diagnostics_agent", "remote_intervention_agent",
                               "cgs_diagnostics_ecs_api", "cgs_diagnostics_streamer_api", "cgs_bagger_server"]
    _docker_compose_container_name = []
    _docker_images = {}
    _docker_images["remote_intervention_agent"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/remote_intervention_agent:latest"
    _docker_images["kriya_watchdog"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/remote_intervention_agent:latest"
    _docker_images["cgs_diagnostics_agent"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/cognicept_diagnostics_agent:latest"
    _docker_images["cgs_diagnostics_ecs_api"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/cognicept_diagnostics_api:latest"
    _docker_images["cgs_diagnostics_streamer_api"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/cognicept_diagnostics_api:latest"
    _docker_images["colab_master"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/ros:master"
    _docker_images["colab_description"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/ros:colab-desc"
    _docker_images["cgs_mission_sys"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/mission_system:latest"
    _docker_images["orbitty"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/orbitty:latest"
    _docker_images["cgs_bagger_server"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/cognicept_rosbagger:latest"
    _docker_images["health_aggregator"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/health_aggregator:dev"
    _docker_images["diagnostics_aggregator"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/health_aggregator:dev"
    _docker_images["smartplus_sound"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/smartplus_sound_server:latest"
    _docker_images["map_manager"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/map_manager:latest"
    _docker_images["computer_health_metrics"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/ros:computer_health_metrics"
    _docker_images["slamtec_adapter"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/slamtec_adapter:latest"
    _docker_images["cam_capture"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/ros:camera_capture"

    def configure_containers(object, cfg):
        """
        Loads agent configuration from `COG_AGENT_CONTAINERS`, `COG_AGENT_IMAGES` and `COG_COMPOSE_FILE`

                Parameters:
                        cfg (Configuration): Cognicept configuration
                Returns:
                        None
        """

        if("COG_AGENT_CONTAINERS" in cfg.config and "COG_AGENT_IMAGES" in cfg.config):
            container_names = cfg.config["COG_AGENT_CONTAINERS"].split(";")
            image_names = cfg.config["COG_AGENT_IMAGES"].split(";")
            if(len(image_names) == len(container_names)):
                object._docker_container_names = container_names
                object._docker_images = {}
                i = 0
                for container_name in container_names:
                    object._docker_images[container_name] = image_names[i]
                    i = i + 1
            else:
                print(
                    "`COG_AGENT_CONTAINERS` and `COG_AGENT_IMAGES` do not coincide. Using default.")

            if("COG_ORBITTY_ENABLED" in cfg.config and "COG_ORBITTY_IMAGE" in cfg.config):
                if(bool(cfg.config["COG_ORBITTY_ENABLED"])):
                    object._docker_images["orbitty"] = cfg.config["COG_ORBITTY_IMAGE"]
        else:
            print(
                "Undefined `COG_AGENT_CONTAINERS` or `COG_AGENT_IMAGES`. Using default.")

        if "COG_COMPOSE_FILE" in cfg.config:
            object._compose_dir = os.path.dirname(os.path.expanduser(cfg.config["COG_COMPOSE_FILE"]))
            docker_compose_dict = cfg.get_docker_compose()

            if docker_compose_dict == {}:
                print(f"{bcolors.WARNING}Warning: failed to load docker compose file - {cfg.config['COG_COMPOSE_FILE']} {bcolors.ENDC}")
            else:
                for container_name in docker_compose_dict:
                    object._docker_container_names.append(container_name)
                    object._docker_compose_container_name.append(container_name)
                object._docker_images.update(docker_compose_dict)

    def _get_latest_log_loc(object, args):
        """
        Retrieve path to the last log in `~/.cognicept/agent/logs/` relative to `~/.cognicept/` or `path` specified by args.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        latest_log_loc (str): path to latest log relative to `~/.cognicept/`
        """
        # get latest log location
        latest_log_loc_file_path = os.path.expanduser(
            args.path+"agent/logs/latest_log_loc.txt")
        latest_log_loc = ""
        try:
            with open(latest_log_loc_file_path) as txt_file:
                latest_log_loc_temp = txt_file.readline()
                latest_log_loc_temp = latest_log_loc_temp[:-1]
                latest_log_loc = latest_log_loc_temp.replace(
                    "/$HOME/.cognicept/", "")
                latest_log_loc = latest_log_loc.replace(".cognicept/", "")
        except:
            cgs_agent_status = bcolors.FAIL + "UNKNOWN" + bcolors.ENDC

        return latest_log_loc

    def get_status(object, args):
        """
        Prints status of docker containers listed in `COG_AGENT_CONTAINERS`.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`

        """
        permission_docker_call_result = permission_safe_docker_call(
            docker.from_env)
        if permission_docker_call_result is None:
            return False
        client = permission_docker_call_result
        # check status of cgs_agent
        # get latest log location
        latest_log_loc = object._get_latest_log_loc(args)
        # read latest status and set for display
        file_path = os.path.expanduser(
            args.path+latest_log_loc+"/logDataStatus.json")
        try:
            with open(file_path) as json_file:
                data = json.load(json_file)
                period_since_update = datetime.utcnow(
                ) - dateutil.parser.parse(data["timestamp"])
                if(period_since_update.seconds < 30 and period_since_update.seconds >= 0):
                    cgs_agent_status = bcolors.OKBLUE + \
                        data["message"].upper() + bcolors.ENDC
                else:
                    cgs_agent_status = bcolors.WARNING + "STALE" + bcolors.ENDC

        except:
            cgs_agent_status = bcolors.FAIL + "Error" + bcolors.ENDC

        container_names = object._docker_container_names
        for container_name in container_names:
            print(container_name, end=': ', flush=True)
            try:
                container = client.containers.get(container_name)
                if container.status != "running":
                    print(bcolors.WARNING + "OFFLINE" + bcolors.ENDC)
                else:
                    if(container_name == "cgs_diagnostics_agent"):
                        print(cgs_agent_status)
                    elif(container_name == "remote_intervention_agent"):
                        object._parse_remote_intervention_agent_logs(
                            container.logs(tail=50))
                    else:
                        print(bcolors.OKBLUE + "ONLINE" + bcolors.ENDC)
            except docker.errors.NotFound:
                print(bcolors.FAIL + "CONTAINER NOT FOUND" + bcolors.ENDC)

    def get_last_event(object, args):
        """
        Prints last log in `~/.cognicept/agent/logs/`.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`

        """
        # get latest log location
        latest_log_loc = object._get_latest_log_loc(args)

        # read and display latest log if any
        file_path = os.path.expanduser(args.path+latest_log_loc)
        try:
            print(bcolors.OKBLUE +
                  "Looking for the latest event log..." + bcolors.ENDC)
            # should read only latest logData#.json file and not logDataStatus.json
            list_of_log_files = [fn for fn in glob.glob(file_path + "/*.json")
                                 if not os.path.basename(fn).endswith("logDataStatus.json")]
            latest_log_file = max(list_of_log_files, key=os.path.getctime)

            with open(latest_log_file) as json_file:
                data = json.load(json_file)
                print(bcolors.OKGREEN+"Latest Event Log:" + bcolors.ENDC)
                print(json.dumps(data, indent=4, sort_keys=True))
        except:
            print(bcolors.WARNING + "No event logs present." + bcolors.ENDC)

    def check_docker_image_exists(object, args):
        """
        Checks if this docker image exists.
        
                Parameters:
                        args (str): Input container names.
        """
        image_set = set()
        missing_image = set()

        # If list of agents is not specified, get all container_names
        if(not hasattr(args, 'list') or len(args.list) == 0):
            args.list = object._docker_container_names

        # Based on args.list to get image(s)
        for container_name in args.list:
            try:
                image_set.add(object._docker_images[container_name])
            except KeyError as exp:
                print(bcolors.WARNING + "Container: " + container_name + " not found. Skipping." + bcolors.ENDC)

        success_flag = True
        docker_client = permission_safe_docker_call(docker.from_env)
        for image in image_set:
            try:
                image = docker_client.images.get(image)

            except docker.errors.ImageNotFound:
                missing_image.add(image) #Track image(s) that is/are missing
                success_flag = False

        return success_flag, missing_image

    def _parse_remote_intervention_agent_logs(object, logs):
        """
        Parses logs to find status of remote intervention agent. Prints status.

                Parameters:
                        logs: container logs

        """
        logs_lines = logs.splitlines()
        # parse logs to get current status
        ri_agent_status = {}
        ri_agent_status["AGENT"] = ""
        ri_agent_status["WEBRTC"] = ""
        ri_agent_status["WEBSOCKET"] = ""

        # find latest status of the each module (agent, webrtc, websocket)
        for line in reversed(logs_lines):
            for key, value in ri_agent_status.items():
                if(value != ""):
                    continue
                matches = re.search(
                    '^.*{}:: STATUS:: (?P<status>.*).*$'.format(key), str(line))
                if(matches is not None):
                    ri_agent_status[key] = matches.groups(0)[0]
            if(ri_agent_status["AGENT"] != "" and ri_agent_status["WEBRTC"] != "" and ri_agent_status["WEBSOCKET"] != ""):
                continue

        output_text = bcolors.OKBLUE + "ONLINE" + bcolors.ENDC

        for key, value in ri_agent_status.items():
            if(value == ""):
                # if not found, it was not yet initialized
                output_text = bcolors.WARNING + "NOT INITIALIZED" + bcolors.ENDC
                break
            if(value != "OK"):
                output_text = bcolors.WARNING + key + value + bcolors.ENDC
                break
        print(output_text)

    def _detached_restart(object, args):
        """
        Runs _restart_protocol in detached mode with printing muted.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """

        if os.fork() != 0:
            return
        sys.stdout = open(os.devnull, 'w')
        result = object._restart_protocol(args)
        sys.stdout = sys.__stdout__
        return result

    def _restart_protocol(object, args):
        """
        Stops and starts the containers listed in `COG_AGENT_CONTAINERS`.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """
        object.stop(args)
        result = object.start(args)
        return result

    def restart(object, args):
        """
        Stops and starts the containers listed in `COG_AGENT_CONTAINERS`.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded in attached mode, always True in detached mode
        """
        success_flag, missing_images = object.check_docker_image_exists(args) #success_flag is a bool to check if all image(s) exist before running restart protocol

        if success_flag:
            if args.detach:
                result = True
                print("Running restart in detached mode")
                p = Process(target=object._detached_restart, args=(args,))
                p.start()
            else:
                result = object._restart_protocol(args)
            if args.prune:
                print("Clearing logs")
                object.clear_logs(args)
                            
        else:
            result = success_flag
            print(bcolors.FAIL + "Error: The following image(s) shown below cannot be found" + bcolors.ENDC)
            print(*missing_images, sep = "\n")
        return result

    def start(object, args):
        """
        Starts the containers listed in `COG_AGENT_CONTAINERS`. If `args` has parameter `list`, starts only containers in the list.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """
        print("Called start")
        print("Starting agents")
        result = object.run_agents(args)
        return result

    def stop(object, args):
        """
        Stops the containers listed in `COG_AGENT_CONTAINERS`. If `args` has parameter `list`, stops only containers in the list.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """
        print("Stopping agents")
        result = object.remove_agents(args)
        return result
    
    def check_sudo_password(object, password):
        '''
        Checks if user entered password for clearing logs is correct. Runs a dummy sudo command to test input password.
        '''
        cmd = ["sudo","-kS","echo","test"]
        
        p = subprocess.run(cmd, capture_output=True, input=password, encoding="ascii")

        if "incorrect password" in str(p):
            return False
        else:
            return True
        
    
    def clear_logs(object, args):
        """
        Remove all unused containers, networks, images (both dangling and unreferenced). Removes all .txt logs stored in /kriya_logs
        and .json logs stored in /agents/logs
                
        """
        log_list = ["sudo","-S","-k", "rm"]
        kriya_log_count = 0
        agent_folder_list = ["sudo","-S","-k","rm","-r"]
        cognicept_dir_path = os.path.expanduser(args.path)
        kriya_logs_dir = cognicept_dir_path + "kriya_logs/"
        if os.path.exists(kriya_logs_dir):
            kriya_logs = os.listdir(kriya_logs_dir)
            for item in kriya_logs:
                if item.endswith(".txt"):
                    log_list.append(str(kriya_logs_dir + item))
                    kriya_log_count += 1

        agent_logs_dir = cognicept_dir_path + "agent/logs/"
        if os.path.exists(agent_logs_dir):
            agent_folders = os.listdir(agent_logs_dir)
            
            for folder in agent_folders:
                folder_path = agent_logs_dir + folder
                if os.path.isdir(folder_path) and folder != "bunched_logs" and folder != "unittest_logs":
                    agent_folder_list.append(str(folder_path))
                
                elif folder == "bunched_logs":
                    bunched_logs = os.listdir(folder_path)
                    bunched_log_count = len(bunched_logs)
                    for item in bunched_logs:
                        log_list.append(str(folder_path + "/" + item))
                
                elif folder == "unittest_logs":
                    unittest_logs = os.listdir(folder_path)
                    unittest_log_count = len(unittest_logs)
                    for item in unittest_logs:
                        log_list.append(str(folder_path + "/" + item))

        agent_folder_count = len(agent_folder_list)-4

        print(f'\nClear File Summary: \n-Kriya: {kriya_log_count} Logs Found \n-Agent: {agent_folder_count} Folders Found \n-Bunched: {bunched_log_count} Logs Found \n-Unittest: {unittest_log_count} Logs Found')
        if kriya_log_count + bunched_log_count + unittest_log_count + agent_folder_count == 0:
            print("No detected Logs and Files to be cleared.")
        else:
            print("Do you wish to proceed?")
            user_input = input("(Y/N)")
            if user_input.lower() == "y":
                
                attempt = 0
                success = False
                while attempt < 3 and success == False:
                    password = getpass.getpass("Enter sudo password: ")
                    success = object.check_sudo_password(password)
                    
                    if not success:
                        
                        attempt += 1
                
                if success:
                
                    if kriya_log_count > 0:
                        print("Successfully removed all kriya_folder")
                    
                    if agent_folder_count > 0:
                        clear_agent_folder = subprocess.Popen(agent_folder_list, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                        stdout, stderr = clear_agent_folder.communicate(input=(password + '\n').encode())
                        print("Successfully removed all agent_folder")

                    if bunched_log_count > 0:
                        print("Successfully removed all bunched_logs")

                    if unittest_log_count > 0:
                        print("Successfully removed all unittest_logs")

                    clear_logs = subprocess.Popen(log_list, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                    stdout, stderr = clear_logs.communicate(input=(password + '\n').encode())

                else:
                    print("Logs and Files not cleared, Sudo Password Incorrect.")
            else:
                print("Logs and Files not cleared")

    def status(object, args):
        object.get_status(args)
        object.status_datadog(args)
        return True

    def remove_agents(object, args):
        """
        Stops the containers listed in `args.list`.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """

        # if list of agents is not specified, restart all
        if(not hasattr(args, 'list') or len(args.list) == 0):
            args.list = object._docker_container_names
        permission_docker_call_result = permission_safe_docker_call(
            docker.from_env)
        if permission_docker_call_result is None:
            return False

        client = permission_docker_call_result
        print("STOP: ")
        flag_success = True
        for container_name in args.list:
            print("   - " + container_name, end=': ', flush=True)
            try:
                container = client.containers.get(container_name)
                container.stop(timeout=10)
                container.remove()
                print(bcolors.OKBLUE + "DONE" + bcolors.ENDC)
            except docker.errors.NotFound:
                print(bcolors.WARNING + "NOT FOUND" + bcolors.ENDC)
                flag_success = False
            except docker.errors.APIError:
                print(bcolors.FAIL + "ERROR" + bcolors.ENDC)
                flag_success = False
        return flag_success

    def run_agents(object, args):
        """
        Starts the containers listed in `args.list`.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """
        # if list of agents is not specified, restart all and run the docker-compose file
        if(not hasattr(args, 'list') or len(args.list) == 0):
            args.list = object._docker_container_names
        
        object._agent_run_options = {}

        # diagnostics agent/API run config
        object._agent_run_options["cgs_diagnostics_agent"] = {"command": "rosrun error_resolution_diagnoser error_resolution_diagnoser", "volumes": {
            args.config.config_path + "agent/logs/": {"bind": "/root/.cognicept/agent/logs", "mode": "rw"}}, "network_mode": "host"}
        object._agent_run_options["cgs_diagnostics_ecs_api"] = {
            "command": "/src/ecs_endpoint.py", "network_mode": "host"}
        object._agent_run_options["cgs_diagnostics_streamer_api"] = {"command": "/src/streamer_endpoint.py", "volumes": {
            args.config.config_path + "agent/logs/bunched_logs": {"bind": "/root/.cognicept/agent/logs/bunched_logs", "mode": "rw"}}, "network_mode": "host"}

        # health aggregator run config
        object._agent_run_options["health_aggregator"] = {"command": "roslaunch health_monitoring_aggregator aggregator.launch --wait", "volumes": {
            args.config.config_path + "health_config.yaml": {"bind": "/home/aggregator_ws/src/health_monitoring_aggregator/config.yml", "mode": "rw"}}, "network_mode": "host"}
        
        # diagnostics aggregator run config
        object._agent_run_options["diagnostics_aggregator"] = {"command": "roslaunch diagnostic_aggregator aggregator.launch --wait", "volumes": {
            args.config.config_path + "diag_analyzers.yaml": {"bind": "/opt/ros/noetic/share/diagnostic_aggregator/demo/pr2_analyzers.yaml", "mode": "rw"}}, "network_mode": "host"}

        # kriya run config
        object._agent_run_options["remote_intervention_agent"] = {
            "command": "", "network_mode": "host"}
        if(args.config.is_ssh_enabled()):
            object._agent_run_options["remote_intervention_agent"]["volumes"] = {
                args.config.config_path + "ssh/id_rsa": {"bind": "/root/.ssh/id_rsa", "mode": "rw"}}
        if(args.config.is_audio_enabled()):
            object._agent_run_options["remote_intervention_agent"]["devices"] = ["/dev/snd"]
        object._agent_run_options["kriya"] = object._agent_run_options["remote_intervention_agent"]
        object._agent_run_options["kriya_watchdog"] = {
            "command": "python3 /home/watchdog.py", "volumes": {
                args.config.config_path + "kriya_logs/": {"bind": "/root/logs/", "mode": "rw"},
                "/var/run/docker.sock/": {"bind": "/var/run/docker.sock/", "mode": "rw"}}}

        # kopilot run config
        object._agent_run_options["kopilot"] = {"command": "", "network_mode": "host", "volumes": {
            args.config.config_path + "kockpit.yaml": {"bind": "/root/config/kockpit.yaml", "mode": "rw"}}}

        # ROSBagger run config
        object._agent_run_options["cgs_bagger_server"] = {"command": "rosrun cognicept_rosbagger bagger_action_server.py", "volumes": {
            args.config.config_path + "bags/": {"bind": "/root/.cognicept/bags", "mode": "rw"}}, "network_mode": "host"}
        
        # Smart+ Sound Server run config
        object._agent_run_options["smartplus_sound"] = {"volumes": {
            args.config.config_path + "sounds/": {"bind": "/root/.cognicept/sounds", "mode": "rw"},
            args.config.config_path + "tts_configuration.yaml": {"bind": "/home/smartplus_sound_server_ws/src/tts/config/sample_configuration.yaml", "mode": "rw"}},
            "network_mode": "host", "devices": ["/dev/snd"]}

        # Map Manager run config
        object._agent_run_options["map_manager"] = {"volumes": {
            args.config.config_path + "slamware_maps/": {"bind": "/root/.cognicept/slamware_maps", "mode": "rw"},
            args.config.config_path + "map_server_maps/": {"bind": "/root/.cognicept/map_server_maps", "mode": "rw"},
            args.config.config_path + "building_info.json": {"bind": "/home/map_manager_ws/src/map_manager/config/sample_building_info.json", "mode": "rw"}},
            "network_mode": "host"}
        
        # Computer Health Metrics run config
        object._agent_run_options["computer_health_metrics"] = {"volumes": {
            args.config.config_path + "health_config.yaml": {"bind": "/home/health_ws/src/computer_health_metrics/config/computer_params.yml", "mode": "rw"},
            "/var/run/docker.sock/": {"bind": "/var/run/docker.sock/", "mode": "rw"}},
             "network_mode": "host"}

        # Slamware adapter run config
        object._agent_run_options["slamtec_adapter"] = {"volumes": {
            args.config.config_path + "slamware_maps/": {"bind": "/root/.cognicept/slamware_maps", "mode": "rw"}},
            "network_mode": "host"}
        
        # Camera capture run config 
        object._agent_run_options["cam_capture"] = {"volumes": {
            args.config.config_path + "inspect_images/" : {"bind": "/root/.cognicept/inspect_images/", "mode": "rw"},
            args.config.config_path + "runtime.env" : {"bind": "/root/.cognicept/runtime.env", "mode": "rw"}},
            "network_mode": "host"}
        

        # Default other config
        object._agent_run_options["other"] = {
            "command": "", "network_mode": "host"}
        permission_docker_call_result = permission_safe_docker_call(
            docker.from_env)
        if permission_docker_call_result is None:
            return False
        client = permission_docker_call_result
        print("RUN: ")
        success_flag = True
        for container_name in args.list:
            print("   - " + container_name, end=': ', flush=True)
            try:
                if(container_name not in object._docker_images):
                    if("COG_AGENT_CONTAINERS" in args.config.config):
                        containers = " (configured list: " + \
                            args.config.config["COG_AGENT_CONTAINERS"] + ")"
                    else:
                        containers = ""
                    print(bcolors.WARNING + "NOT FOUND" +
                          bcolors.ENDC + containers)
                    success_flag = False
                    continue

                if(container_name in object._agent_run_options.keys()):
                    options = object._agent_run_options[container_name]
                else:
                    options = object._agent_run_options["other"]
                options["name"] = container_name
                options["detach"] = True
                options["environment"] = args.config.config
                try:
                    if container_name == "smartplus_sound":
                        options["environment"]["ALSA_CARD"] = options["environment"]["SOUND_DEV_OUT"]
                    elif container_name == "remote_intervention_agent":
                        options["environment"]["ALSA_CARD"] = options["environment"]["SOUND_DEV_IN"]
                except KeyError as exp:
                    print(bcolors.WARNING + 'Missing sound device configuration in runtime.' +
                        ' Might result in sound features not working as expected.' +
                        ' Explicitly define `SOUND_DEV_IN` and `SOUND_DEV_OUT` variables.' + 
                        bcolors.ENDC)
                options["restart_policy"] = {"Name": "unless-stopped"}
                options["tty"] = True
                options["log_config"] = LogConfig(
                    type=LogConfig.types.JSON, config={'max-size': '5m'})
                if "command" in options:
                    command = options.pop("command")
                else:
                    command = ""
                
                #if container name is within docker-compose service, trigger run function
                # check whether the container's image is not found
                if container_name in object._docker_compose_container_name:
                    object._run_docker_compose_service(container_name)
                else:
                    container = client.containers.run(
                        object._docker_images[container_name], command, **options)
                
                print(bcolors.OKBLUE + "DONE" + bcolors.ENDC)
            except docker.errors.ContainerError:
                print(bcolors.WARNING + "ALREADY EXISTS" +
                      bcolors.ENDC + " (run `cognicept update`)")
            except docker.errors.ImageNotFound:
                print(bcolors.WARNING + "IMAGE NOT FOUND" +
                      bcolors.ENDC + " (run `cognicept update`)")
                success_flag = False
            except docker.errors.APIError as exp:
                print(exp)
                print(bcolors.FAIL + "DOCKER ERROR" + bcolors.ENDC)
                success_flag = False
        return success_flag

    def status_datadog(object, args):
        """
        Prints datadog status.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """
        sts = subprocess.call(
            ['sudo', 'sh', '-c', "systemctl status datadog-agent", "-p"], shell=True, stdout=DEVNULL, stderr=DEVNULL)
        if sts == 0:
            print("Health monitor status:" +
                  bcolors.OKBLUE + " ACTIVE" + bcolors.ENDC)
        elif sts == 4:
            print("Health monitor status:" +
                  bcolors.FAIL + " NOT FOUND" + bcolors.ENDC)
        else:
            print("Health monitor status:" +
                  bcolors.WARNING + " INACTIVE" + bcolors.ENDC)

    def run_orbitty(object, args):
        """
        Starts Orbitty.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        None
        """
        os.system("xhost +local:root")
        permission_docker_call_result = permission_safe_docker_call(
            docker.from_env)
        if permission_docker_call_result is None:
            return False
        client = permission_docker_call_result
        try:
            options = {}
            options["name"] = "orbitty"
            options["detach"] = False
            options["privileged"] = True
            options["volumes"] = {}
            options["volumes"][args.config.config_path] = {
                "bind": "/config", "mode": "rw"}
            options["volumes"]["/tmp/.X11-unix"] = {
                "bind": "/tmp/.X11-unix", "mode": "rw"}
            environment = args.config.config
            environment["QT_X11_NO_MITSHM"] = 1
            environment["DISPLAY"] = ":0"
            options["environment"] = args.config.config
            options["remove"] = True
            options["tty"] = True
            command = "roslaunch orbitty orbitty.launch"
            client.containers.run(
                object._docker_images["orbitty"], command, **options)
        except docker.errors.ContainerError:
            print(bcolors.WARNING + "ALREADY RUNNING" + bcolors.ENDC)
        except docker.errors.ImageNotFound:
            print(bcolors.WARNING + "IMAGE NOT FOUND" +
                  bcolors.ENDC + " (run `cognicept update`)")
        except docker.errors.APIError:
            print(bcolors.FAIL + "DOCKER ERROR" + bcolors.ENDC)
        os.system("xhost -local:root")

    def cognicept_version_update(object):
        current_version = pkg_resources.require("cognicept-shell")[0].version
        package = 'cognicept-shell'
        response = requests.get(f'https://pypi.org/pypi/{package}/json')
        latest_version = response.json()['info']['version']
        if latest_version != current_version:
            print(f"{package} current version {current_version} - Installing Version {latest_version}")
            os.system(f'pip3 install -q {package}=={latest_version}')
            print(f'Installation {package} to version {latest_version}:'+bcolors.OKBLUE + " DONE" + bcolors.ENDC)
        else:
            print(f"{package} already in latest version={latest_version}")

    def pull_image(object,image_name,N,i):
        """
        Pulls a Docker image and displays a progress bar.

            Parameters:
                image_name : The name of the Docker image to pull.
                N : The total number of images to pull.
                i : index
        """
        image_name_short = image_name.replace("412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/", "cognicept/")
        for status in object.docker_client.pull(image_name, stream=True, decode=True):
            if("progress" not in status):
                status["progress"] = ""
            if("status" not in status):
                status["status"] = "Error"
            terminal_size = shutil.get_terminal_size().columns
            progress = ""
            if("progressDetail" in status and "total" in status["progressDetail"]):
                progress = generate_progress_bar(
                    status["progressDetail"]["current"], status["progressDetail"]["total"], 1, 10)
                status = "[" + str(i) + "/" + str(N) + "] " + image_name_short + \
                    " - " + status["status"] + " " + progress
                if(terminal_size > 0):
                    print('{:{terminal_size}.{terminal_size}}'.format(
                        status, terminal_size=terminal_size), end="\r", flush=True)
                else:
                    print('{:{trm_sz}.{trm_sz}}'.format(
                        status, trm_sz=80), end="\r", flush=True)
        print("[" + str(i) + "/" + str(N) + "] " + image_name_short +
            " - " + bcolors.OKBLUE + "OK" + bcolors.ENDC + "\033[K")

    def update_agents(object, args):
        """
        Starts the containers listed in `args.list` for the purpose of update.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """
        
        # if list of agents is not specified, update all agents
        
        if((not hasattr(args, 'list') or not args.list) and (not hasattr(args, 'image') or not args.image)):
            images = set(object._docker_images.values())
            # load extra images to update
            if("COG_EXTRA_IMAGES" in args.config.config):
                image_names = args.config.config["COG_EXTRA_IMAGES"].split(";")
                if(len(image_names) > 0):
                    images = images.union(set(image_names))
            
            N = len(images)
            object.cognicept_version_update()
            print("Info: This may take a while depending on your connection.")
        # if list of agents is specified, update only that particular set of agents
        else: 
            images = set()
            if args.list:
                for container_name in args.list:
                    if container_name in object._docker_container_names:
                        images.add(object._docker_images[container_name])
                    else:
                        print(f"Error: Container {container_name} not found")
                        success_flag = False

            if args.image:
                for image_name in args.image:
                    images.add(image_name)

            N = len(images)
            print("Info: Update " + str(N) + " agent_image(s).")

        i = 0 # For indexing
        success_flag = True

        for image_name in images:
            i = i + 1
            try:
                if('412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/' in image_name):
                    object.pull_image(image_name, N=N, i=i)
                else:
                    object.pull_image('412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/' + image_name, N=N, i=i)

            except docker.errors.NotFound:
                try:
                    object.pull_image(image_name,N=N, i=i)
                    
                except docker.errors.NotFound:
                    print(bcolors.FAIL + "Error: " + bcolors.ENDC + f"Image {image_name} can’t be accessed in ECR or Docker Hub.")
                    print("[" + str(i) + "/" + str(N) + "] " + image_name +
                        " - " + bcolors.FAIL + "FAILED" + bcolors.ENDC + "\033[K")
                    success_flag = False
                    
                except DockerException as ex:
                    print("[" + str(i) + "/" + str(N) + "] " + image_name +
                        " - " + bcolors.FAIL + "FAILED" + bcolors.ENDC + "\033[K")
                    print(f"Error: {ex}")
                    success_flag = False
                    continue  # skip the current image and continue with the next one
            except:
                print("[" + str(i) + "/" + str(N) + "] " + image_name +
                    " - " + bcolors.FAIL + "FAILED" + bcolors.ENDC + "\033[K")
                success_flag = False

        if not success_flag:
            print("There were errors while updating some images.")
        print("Info: Run `cognicept restart` to redeploy updated agents.")

        return success_flag
    
    def _update_protocol(object, args):
        """
        Pulls docker images listed in `COG_AGENT_IMAGES`.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """
        #Check internet connection, if cannot connect, notify user and return false
        response_time = ping('google.com')
        if response_time == False:
            print("No internet connection. Please check the internet connection and try 'cognicept update` again.")
            return False
        else:
            login_success = object._ecr_login(args)
            if login_success is False:
                print("Your login has failed due to the error mentioned above. Please try again.")
                return False
            if object.update_agents(args) is False:
                config_file_path = os.path.expanduser('~/.docker/config.json')
                if os.path.exists(config_file_path):
                    print(bcolors.WARNING + "There is a conflict in authentication file, run rm ~/.docker/config.json to remove conflicting file" + bcolors.ENDC)   
                    return False
                else:
                    print(bcolors.WARNING + "Your update credentials may have expired. Run `cognicept keyrotate` to refresh credentials and try `cognicept update` again." + bcolors.ENDC)
                    return False   
        
    def _detached_update(object, args):
        """
        Runs the _update_protocol that pulls docker images listed in `COG_AGENT_IMAGES`.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """
        if os.fork() != 0:
            return

        sys.stdout = open(os.devnull, 'w')

        result = object._update_protocol(args)

        sys.stdout = sys.__stdout__

        return result

    def update(object, args):
        """
        Pulls docker images listed in `COG_AGENT_IMAGES`. Can be run in detached mode where printing is muted

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): Always True if run in detached mode, otherwise True if successful
        """
        if args.detach:
            result = True
            print("Running update in detached mode")
            p = Process(target=object._detached_update, args=(args,))
            p.start()
        else:
            result = object._update_protocol(args)

        return result

    def _construct_ecr_client(object, config_obj):
        """
        Member utility function to create ECR client `object.ecr_client` based on runtime.env credentials

                Parameters:
                        config_obj (Configuration): object holding Cognicept configuration
                Returns:
                        None      
        """
        # Get config
        local_cfg = config_obj.fetch_aws_keys()
        if local_cfg == False:
            return False
        if 'SessionToken' in local_cfg:
            object.ecr_client = boto3.client(
                'ecr', region_name='ap-southeast-1',
                aws_access_key_id=local_cfg['AccessKeyId'],
                aws_secret_access_key=local_cfg['SecretAccessKey'],
                aws_session_token=local_cfg['SessionToken'])
        else:
            object.ecr_client = boto3.client(
                'ecr', region_name='ap-southeast-1',
                aws_access_key_id=local_cfg['AccessKeyId'],
                aws_secret_access_key=local_cfg['SecretAccessKey'])
        return True

    def _ecr_login(object, args):
        """
        Member utility function that is called to login to ECR

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """
        # Construct client
        result = object._construct_ecr_client(args.config)
        if result == False:
            return False
        num_retries = 3
        for trial in range(num_retries):
            try:
                # Get token
                token = object.ecr_client.get_authorization_token()
                # Parse username, password
                username, password = base64.b64decode(
                    token['authorizationData'][0]['authorizationToken']).decode().split(':')
                # Get registry
                registry = token['authorizationData'][0]['proxyEndpoint']
                # Login
                docker_call_result = permission_safe_docker_call(docker.APIClient,
                                                                 base_url='unix://var/run/docker.sock')
                if docker_call_result is None:
                    return False

                object.docker_client = docker_call_result

                object.docker_client.login(username, password,
                                           registry=registry, reauth=True)
                # Return true for successful login
                return True
            except (docker.errors.APIError, botocore.exceptions.ClientError) as e:
                # On failure, retry
                print('Attempt #' + str(trial+1) + bcolors.FAIL +
                      " FAILED" + bcolors.ENDC + "\033[K")
                error_message = str(e)
                # Wait for 1 second before retrying
                time.sleep(1.0)
        # If the loop is completed, login failed, so return false
        print(bcolors.FAIL + error_message + bcolors.ENDC)
        return False

    def display_version(object, args):
        """
        Display Cognicept-Shell version and docker images version

        Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
        """
        version = pkg_resources.require("cognicept-shell")[0].version
        data = {}
        images_version = []
        data['Container Name'] = object._docker_container_names
        # default value for version: "unknown/latest"
        for x in data['Container Name']:
            if x in object._docker_images:
                image_data = object._docker_images[x].split(':')
                if len(image_data) > 1:
                    images_version.append(image_data[1])
                    image_data = []
            else:
                images_version.append("unknown/latest")
        data['Version'] = images_version
        print("Cognicept Shell Version "+version)
        print(tabulate(data, headers='keys', tablefmt='psql'))
        print("Runtime enviroment file directory: " +args.path)


    def _run_docker_compose_service(object, container_name,):
        option = {"--detach":True, 
                "SERVICE" :[f"{container_name}"],
                "--no-deps": "",
                "--no-color":True,
                "--no-recreate": True,
                "--always-recreate-deps": "",
                "--abort-on-container-exit": False,
                "--remove-orphans": False,
                "--force-recreate": False,
                "--build": False,
                "--no-build": True,
                "--scale": [f'{container_name}=1'],
                "--quiet-pull": True,
                }
        project = project_from_options(object._compose_dir,option)
        compose_cmd = TopLevelCommand(project)
        compose_cmd.up(option)




# coding=utf8
# Copyright 2020 Cognicept Systems
# Author: Jakub Tomasek (jakub@cognicept.systems)
# --> Configuration class handles Cognicept configuration

from dotenv import dotenv_values
from pathlib import Path
from Crypto.PublicKey import RSA
import subprocess
import os
import re
import jwt
import requests
import getpass
import errno
import shutil
import yaml

from cogniceptshell.common import bcolors


class Configuration:
    """
    A class to manage agent and Cognicept's configuration
    ...

    Parameters
    ----------
    None

    Methods
    -------    
    load_config(path):
        Loads and parses `runtime.env` file into `object.config` located at `path`.
    configure(args):
        Manages cognicept configuration based on values in `args`.
    save_config():
        Saves configuration into `runtime.env` file.
    get_cognicept_credentials():
        Checks and returns `COGNICEPT_ACCESS_KEY` from configuration.
    get_cognicept_api_uri():
        Checks and returns `COGNICEPT_API_URI` from configuration.
    get_cognicept_user_api_uri():
        Checks and returns `COGNICEPT_USER_APT_URI` from configuration.
    get_field():
        Returns single config value
    configure_ssh():
        Function to setup ssh access for the host machine given user input.
    is_ssh_enabled(args):
        Checks and returns value of `COG_ENABLE_SSH_KEY_AUTH` as bool.
    is_audio_enabled(args):
        Checks and returns value of `COG_ENABLE_AUDIO` as bool
    cognicept_key_rotate:
        Rotates AWS temporary keys retriving them from Cognicept API.
    init_config(args):
        Initiates runtime.env file with values from Cognicept API
    """
    config_path = os.path.expanduser("~/.cognicept/")
    env_path = config_path + "runtime.env"
    _regex_key = r"^([_A-Z0-9]+)$"
    _config_loaded = False

    def load_config(object, path):
        """
        Loads and parses `runtime.env` file into `object.config` located at `path`.

                Parameters:
                        path (str): Cognicept path, e.g. `~/.cognicept/`.                
        """
        object.config_path = os.path.expanduser(path)
        object.env_path = object.config_path + "runtime.env"
        file = Path(object.env_path)

        if ((not file.exists()) or (file.is_dir())):
            print("Configuration file `" + object.env_path + "` does not exist.")
            try:
                with open(object.env_path, 'w') as f:
                    pass
            except:
                print("Failed to initialize the robot: insufficient priviledges to create `/home/username/.cognicept/runtime.env file.")
                return False

        #object.config = dotenv_values(dotenv_path=file.name) if sys.version_info.minor > 5 else dotenv_values(dotenv_path=find_dotenv(), verbose=True)
        object.config = dotenv_values(
            dotenv_path=object.env_path, verbose=True)
        if(len(object.config) == 0):
            print("Configuration file `" + object.env_path +
                  "` is empty or could not be parsed.")
            return False
        object._config_loaded = True
        return True

    def get_docker_compose(object):
        if "COG_COMPOSE_FILE" in object.config:
            compose_file = os.path.expanduser(object.config["COG_COMPOSE_FILE"])
            compose_images = {}
            try:
                with open(compose_file, 'r') as stream:
                    docker_compose=yaml.safe_load(stream)
                    for container_name in docker_compose['services']:
                        compose_images[container_name] = docker_compose['services'][container_name]['image'] 
                return compose_images
            except (yaml.YAMLError,FileNotFoundError, IOError, TypeError):
                return {}
        else:
            return {}

    def configure(object, args):
        """
        Manages cognicept configuration based on values in `args`:
            * If `args.read` is True, then it prints all configuration,
            * If `args.add` is True, then it will add or modify a single value inputed by user,
            * Otherwise it will iterate through all values and asks for modifications.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`            
        """
        if(args.ssh):
            if object.configure_ssh():
                print(
                    'SSH config done. To apply changes restart agents using `cognicept restart`.')
        elif(args.autocomplete):
            if object.configure_autocomplete():
                print('Autocomplete setup done')
        else:
            if(not object._config_loaded):
                return

            if (not os.access(object.env_path, os.W_OK)):
                print("Error: You don't have writing permissions for `" +
                    object.env_path + "`. Run as `sudo` or change file permissions.")
                return
            if(args.read):
                for key, value in object.config.items():
                    print(key + ': "' + value + '"')
            elif(args.add):
                new_key = ""
                while(new_key == ""):
                    new_key = input("Config name: ")

                    # if empty, exit
                    if(new_key == ""):
                        return
                    # check if matches key specs
                    matches = re.match(object._regex_key, new_key)
                    if matches is None:
                        print(
                            "Error: Key can be uppercase letters, digits, and the '_'. Try again.")
                        new_key = ""

                new_value = ""
                while(new_value == ""):
                    new_value = input("Value: ")
                    if(new_value == ""):
                        return
                    matches = re.match(r"^.*[\"].*$", new_value)
                    if matches is not None:
                        print("Error: Value cannot contain '\"'. Try again.")
                        new_value = ""

                object.config[new_key] = new_value        
            else:
                for key, value in object.config.items():
                    new_value = input(key + "[" + value + "]:")
                    matches = re.match(r"^.*[\"].*$", new_value)
                    if((new_value != "") and (matches == None)):
                        object.config[key] = new_value
            object.save_config()

    def save_config(object):
        """
        Saves configuration into `runtime.env` file

                Parameters:
                    None
        """
        # Try backing up configuration before saving
        backup_success = False
        try:
            shutil.copyfile(object.env_path, object.env_path+'.bk')
            backup_success = True
            print(bcolors.OKBLUE + "Backed up runtime configuration to: " + \
                object.env_path + ".bk" + bcolors.ENDC)
        except OSError as ex:
            if ex.errno == errno.ENOSPC:
                # Catch out of space exceptions
                print(bcolors.FAIL + "Could not back up runtime configuration." + \
                    " Aborting saving runtime configuration." + \
                    " No space left on device!" + bcolors.ENDC)
            else:
                # Raise all other exceptions
                print(bcolors.FAIL + "Could not back up runtime configuration." + \
                    " Aborting saving runtime configuration: " + ex.strerror + bcolors.ENDC)
        
        # If backup is success, try to save it
        if backup_success:
            try:
                with open(object.env_path, 'w') as file:
                    for key, value in object.config.items():
                        file.write(key + '=' + value + '\n')
            except OSError:
                print(bcolors.FAIL + "Could not write into `" + object.env_path +
                    "`. Please check write permission or run with `sudo`." + bcolors.ENDC)        

    def get_cognicept_credentials(object):
        """
        Checks and returns `COGNICEPT_ACCESS_KEY` from configuration.

                Returns:
                    cognicept_access_key (str): value of `COGNICEPT_ACCESS_KEY`
        """
        if "COGNICEPT_ACCESS_KEY" in object.config:
            return object.config["COGNICEPT_ACCESS_KEY"]
        else:
            print('COGNICEPT_ACCESS_KEY missing')

    def get_cognicept_api_uri(object):
        """
        Checks and returns `COGNICEPT_API_URI` from configuration.

                Returns:
                    cognicept_api_uri (str): value of `COGNICEPT_API_URI`, defaults to "https://app.cognicept.systems/api/v1/"
        """
        if "COGNICEPT_API_URI" in object.config:
            return object.config["COGNICEPT_API_URI"]
        else:
            return "https://app.cognicept.systems/api/v1/"

    def get_cognicept_user_api_uri(object):
        """
        Checks and returns `COGNICEPT_USER_API_URI` from configuration.

                Returns:
                    cognicept_user_api_uri (str): value of `COGNICEPT_USER_API_URI`, defaults to "https://app.cognicept.systems/api/v1/"
        """
        if "COGNICEPT_USER_API_URI" in object.config:
            return object.config["COGNICEPT_USER_API_URI"] 
        else:
            return "https://app.cognicept.systems/api/v1/"

    def get_field(object, field_name):
        """
        Returns single config value

                Parameters:
                        field_name (str): name of config field
                Returns:
                        value (str): value
        """
        if field_name in object.config:
            return object.config[field_name]
        else:
            raise KeyError(field_name)

    def _interpret_bool_input(object, input_string):
        """
        Parses user input string bool value 

                Parameters:
                        input_string (str): input value
                Returns:
                        bool_value (bool): True if value is 'Y'
        """
        if(input_string == 'Y'):
            return True
        elif(input_string == 'n'):
            return False
        else:
            return None

    def configure_ssh(object):
        """
        Function to setup ssh access for the host machine given user input

                Parameters:
                        None
                Returns:
                        result (bool): True if some step didn't fail
        """
        print('SSH is used to access the host machine from the isolated docker environment of Cognicept agent.')

        enable_ssh = None
        while(enable_ssh == None):
            enable_ssh = object._interpret_bool_input(
                input("Enable SSH access? (Y/n):"))

        if(not enable_ssh):
            object.config["COG_ENABLE_SSH"] = "False"
            object.config["COG_ENABLE_SSH_KEY_AUTH"] = "False"
            object.config["COG_ENABLE_AUTOMATIC_SSH"] = "False"
            object.save_config()
            return True
        else:
            object.config["COG_ENABLE_SSH"] = "True"

        ssh_authorized_keys_path = os.path.expanduser(
            "~") + "/.ssh/authorized_keys"
        print("\n \nSSH key needs to be used for hosts with disabled password login. "
              "It can also simplify access so you don't need to input the password each time.\n"
              "This process will generate ssh key locally and mount it to the docker container. "
              "The public key is copied to `" + ssh_authorized_keys_path +
              "` to give access.  Root access is needed and you will be prompted for password."
              "The ssh key is neither sent nor stored to the Cognicept server. "
              "If you choose not to, manual password ssh access can be still used.")

        enable_ssh_key = None
        while(enable_ssh_key == None):
            enable_ssh_key = object._interpret_bool_input(
                input("Generate SSH key and give access? (Y/n):"))

        if(not enable_ssh_key):
            object.config["COG_ENABLE_SSH_KEY_AUTH"] = "False"
        else:
            object.config["COG_ENABLE_SSH_KEY_AUTH"] = "True"

        # generate the ssh key and write them in the file
        if object.config["COG_ENABLE_SSH_KEY_AUTH"] == "True":
            cognicet_ssh_directory = object.config_path + "ssh/"
            try:
                if not os.path.exists(cognicet_ssh_directory):
                    os.makedirs(cognicet_ssh_directory)
            except:
                print(
                    "Failed. Don't have privileges to create files/directories within the ssh directory.")
                return False

            # generate the keys
            ssh_key = RSA.generate(2048)
            private_key_path = cognicet_ssh_directory + "id_rsa"
            public_key_path = cognicet_ssh_directory + "id_rsa.pub"
            config_file_path = cognicet_ssh_directory + "config"
            with open(private_key_path, 'wb') as content_file:
                os.chmod(private_key_path, 0o600)
                content_file.write(ssh_key.exportKey('PEM'))
            pub_key = ssh_key.publickey().exportKey('OpenSSH')
            with open(public_key_path, 'wb') as content_file:
                content_file.write(pub_key)
            # add new line at the end of the file
            with open(public_key_path, 'a') as content_file:
                content_file.write("\n")

        default_user = getpass.getuser()
        user_exists = False
        ssh_directory_path = ""

        # retrieve the user name
        user_prompt_number = 0
        while not user_exists:
            user_name = input(
                "Name of the user to access ssh(if empty, defaults to `" + default_user + "`): ")
            if(user_name == ""):
                user_name = default_user
            object.config["COG_SSH_DEFAULT_USER"] = user_name
            
            if(user_name == "root"):
                ssh_directory_path = "/root/.ssh/"
            else:                
                ssh_directory_path = "/home/" + user_name + "/.ssh/"
            if os.path.exists(ssh_directory_path):
                user_exists = True
            else:
                print("User " + user_name +
                      " doesn't seem to exist or openssh server is not installed.")
                user_prompt_number = user_prompt_number + 1
                if(user_prompt_number > 2):
                    print("Failed to find user's openssh configuration directory.")
                    return False

        # copy the keys to authorized_keys file if automatic authentication was enabled
        if object.config["COG_ENABLE_SSH_KEY_AUTH"] == "True":
            authorized_keys_path = ssh_directory_path + "authorized_keys"
            try:
                print('Root access is needed to modify `' +
                      authorized_keys_path + '` and you will be prompted for password.')
                proc = subprocess.call(
                    ['sudo', 'sh', '-c', 'cat ' + public_key_path + ' >> ' + authorized_keys_path])
            except:
                print("Failed! Don't have access to " + authorized_keys_path)
                return

        enable_automatic_ssh = None
        while(enable_automatic_ssh == None):
            enable_automatic_ssh = object._interpret_bool_input(
                input("Enable automatic ssh access? (Y/n):"))

        if(not enable_automatic_ssh):
            object.config["COG_ENABLE_AUTOMATIC_SSH"] = "False"
        else:
            object.config["COG_ENABLE_AUTOMATIC_SSH"] = "True"

        object.save_config()
        return True

    def is_ssh_enabled(object):
        """
        Checks and returns value of `COG_ENABLE_SSH_KEY_AUTH` as bool

                Parameters:
                        None
                Returns:
                        ssh_enabled (bool): True if ssh is enabled
        """
        if(not "COG_ENABLE_SSH_KEY_AUTH" in object.config):
            return False

        return object.config["COG_ENABLE_SSH_KEY_AUTH"]

    def is_audio_enabled(object):
        """
        Checks and returns value of `COG_ENABLE_AUDIO` as bool

                Parameters:
                        None
                Returns:
                        sound enabled (bool): True if sound is enabled
        """
        if(not "COG_ENABLE_AUDIO" in object.config):
            return False

        return object.config["COG_ENABLE_AUDIO"]
    

    def fetch_aws_keys(object):
        """
        Fetch AWS temporary keys retriving them from Cognicept API.

                Parameters:
                        
                Returns:
                        result (json): AWS temporary credentials returned from Cognicept API
        """
        try:
            headers = {
                'Authorization': 'Basic ' + object.get_cognicept_credentials()
            }
            resp = requests.get(object.get_cognicept_api_uri(
            ) + 'aws/assume_role', headers=headers, timeout=5)
            if resp.status_code != 200:
                print('Login error: wrong credentials.')
                return False
            return resp.json()            
        except requests.exceptions.Timeout:
            print("Cognicept REST API error: time out.")
            return False
        except requests.exceptions.TooManyRedirects:
            print("Cognicept REST API error: Wrong endpoint.")
            return False
        except Exception as e:
            print("Cognicept REST API error" + str(e))
            return False
        object.save_config()

        return True

    def cognicept_key_rotate(object, args):
        """
        Rotates AWS temporary keys retriving them from Cognicept API.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """
        print("Updating cloud credentials.")
        try:
            response = object.fetch_aws_keys()
            if response==False:
                print("Error connecting Cognicept API. Please contact support@cognicept.systems")
                return False
            object.config["AWS_ACCESS_KEY_ID"] = response[
                "AccessKeyId"]
            object.config["AWS_SECRET_ACCESS_KEY"] = response[
                "SecretAccessKey"]
            object.config["AWS_SESSION_TOKEN"] = response[
                "SessionToken"]
            object.config["AWS_TOKEN_EXPIRATION"] = response[
                "Expiration"]
            print("Cloud access keys rotated successfully!")
        except requests.exceptions.Timeout:
            print("Cognicept REST API error: time out.")
            return False
        except requests.exceptions.TooManyRedirects:
            print("Cognicept REST API error: Wrong endpoint.")
            return False
        except Exception as e:
            print("Cognicept REST API error" + str(e))
            raise SystemExit()
        object.save_config()

        return True

    
    def init_config(object, args):
        """
        Initiates runtime.env file with values from Cognicept API.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """
        username = input('Username: ')
        password = getpass.getpass('Password: ')
        api_uri = object.get_cognicept_user_api_uri()
        try:
            x = requests.post(api_uri + 'user/login', json={'username': username, 'password': password})
            if 'access_token' not in x.json():
                print('Failed to initialize the robot: Wrong credentials') 
            else:
                auth_key = x.json()["access_token"]
                # check whether whether user activate the 2FA
                decode_json = jwt.decode(auth_key, verify=False, algorithms=["HS256"])

                # False means 2FA Verification is activated by the user
                if decode_json['user_claims']['authorized'] == False:
                    new_key = object.mfa_verfication(api_uri,auth_key)
                    if new_key == auth_key:    
                        print("Failed to initialized robot: Invalid OTP entered 3 times")
                        return False
                    else:
                        auth_key=new_key

                r = requests.get(api_uri + 'spinup/config/' + args.org_id + '/' + args.robot_id, headers={'Authorization': 'Bearer {}'.format(auth_key)})
                j = r.json()
                if j is not None:
                    for key in j:
                        object.config[key] = j[key]
                    object.save_config()
                    print("Successfully initialized configuration for the robot `" + args.robot_id + "`. To start agents run `cognicept start`")
                    return True
                else:
                    print("Failed to initialize the robot: ID `" + args.robot_id + "` in organization `" + args.org_id + "` not found")
                    return False
        except requests.exceptions.Timeout:
            print("Cognicept REST API error: time out.")
            return False
        except requests.exceptions.TooManyRedirects:
            print("Cognicept REST API error: Wrong endpoint.")
            return False
        except Exception as e:
            print("Cognicept REST API error" + str(e))
            raise SystemExit()

    def mfa_verfication(object,api_uri,auth_key):
        otp_trial = 3
        loop = True
        while loop:
            if otp_trial == 0:
                key = auth_key
                loop = False
            else:
                otp = getpass.getpass('OTP from Authenticator: ')
                x = requests.post(api_uri + 'user/mfa/verify', headers={'Authorization': 'Bearer {}'.format(auth_key)}, json={'otp': otp})
                if 'access_token' not in x.json():
                    print('Invalid OTP! Please try again...')
                else:
                    key = x.json()["access_token"]
                    loop = False
            otp_trial -= 1
        return key
             
    def configure_autocomplete(object):
        """

        Function to setup autcomplate functionality
                Parameters:
                        None
                Returns:
                        result (bool): True if some step didn't fail

        """

        enable_autocomplete = None
        while(enable_autocomplete == None):
            enable_autocomplete = object._interpret_bool_input(
                input("Enable Autocomplete? (Y/n):"))
        
        if (not enable_autocomplete):
            print("Autocomplete was not configured")
            return True

        bash_complete_path = "/etc/bash_completion.d/"
        tmp_path = os.path.expanduser("~") + "/.cognicept/"
        
        print("Root access is needed to place bash completition script at `" + bash_complete_path +
              "' you will be prompted for password.")

        allow_root_access = None
        while(allow_root_access == None):
            allow_root_access = object._interpret_bool_input(
            input("Generate script and give access? (Y/n):"))
        subprocess.call(['bash', '-c', 'activate-global-python-argcomplete --dest ' + tmp_path])
        subprocess.call(['sudo', 'bash', '-c', 'mv ' + tmp_path + "python-argcomplete" + " /etc/bash_completion.d/"])
        print("Please restart the terminal.....")
        return True



# PYTHON_ARGCOMPLETE_OK

# Copyright 2020 Cognicept Systems
# Author: Jakub Tomasek (jakub@cognicept.systems)
# --> Main executable file to handle input arguments

import argparse
import argcomplete
import pkg_resources
from cogniceptshell.configuration import Configuration
local_cfg = Configuration()

container_names = []

#added in pull COG_AGENT_IMAGES from "~/.cognicept/"
if(local_cfg.load_config("~/.cognicept/")):
    container_names = local_cfg.config["COG_AGENT_CONTAINERS"].split(";")
    for container_name in local_cfg.get_docker_compose():
        container_names.append(container_name)
    container_names.append([])

#If container_names or agent_image_names cannot be pulled from runtime.env, label preset names for respective sets
else:
    container_names = ["remote_intervention_agent","kriya_watchdog","cgs_diagnostics_ecs_api","cgs_diagnostics_streamer_api","cgs_diagnostics_agent", "colab_master", "colab_description", "cgs_mission_sys", "health_aggregator"]

parser = argparse.ArgumentParser(
        description='Shell utility to configure Cognicept tools.')

parser.add_argument('--version', action='version',
                    version=pkg_resources.require("cognicept-shell")[0].version)

subparsers = parser.add_subparsers(help='', title="Commands")

parser_version = subparsers.add_parser(
    'version', help='Display Cognicept Version')
parser_config = subparsers.add_parser(
    'config', help='Configure Cognicept tools')
parser_status = subparsers.add_parser(
    'status', help='Get status of Cognicept agents')
parser_lastevent = subparsers.add_parser(
    'lastevent', help='Display last event log reported by Cognicept agent')
parser_update = subparsers.add_parser(
    'update', help='Update Cognicept tools')
parser_keyrotate = subparsers.add_parser(
    'keyrotate', help='Rotate Cognicept cloud keys')
parser_restart = subparsers.add_parser(
    'restart', help='Restart Cognicept agents')
parser_start = subparsers.add_parser(
    'start', help='Start Cognicept agents')
parser_stop = subparsers.add_parser('stop', help='Stops Cognicept agents')
parser_orbitty = subparsers.add_parser('orbitty', help='Run Orbitty')
parser_record = subparsers.add_parser(
    'record', help='Manages rosbag recording')
parser_push = subparsers.add_parser(
    'push', help='Pushes stuff to Cognicept cloud')
parser_init = subparsers.add_parser(
    'init', help='Initiate runtime.env')


DEFAULT_PATH = "~/.cognicept/"

parser_version.add_argument(
    '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)
parser_config.add_argument(
    '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)
parser_config.add_argument(
    '--add', help='Add new environmental variable in config file', action='store_true')
parser_config.add_argument(
    '--ssh', help='Configure ssh access during remote intervention', action='store_true')
parser_config.add_argument(
    '--autocomplete', help='Setup autocomplete of cognicept', action='store_true')

parser_config.add_argument(
    '--read', help='Prints Cognicept configuration', action='store_true')


parser_keyrotate.add_argument(
    '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)

parser_status.add_argument(
    '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)

parser_lastevent.add_argument(
    '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)


parser_restart.add_argument('-d', '--detach', action='store_true',
                            help='sets restart to be run in a detached manner. Progress will not be printed')
parser_restart.add_argument(
    '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)
parser_restart.add_argument(
    'list', help='List of agents to restart, leave empty to restart all agents', metavar='list', type=str, nargs='*',choices=container_names)
parser_restart.add_argument('--prune', action='store_true',
                            help='Clears the logs of in kriya_logs and agent logs')


parser_start.add_argument(
    '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)
parser_start.add_argument(
    'list', help='List of agents to start, leave empty to start all agents', metavar='list', type=str, nargs='*',choices=container_names)


parser_stop.add_argument(
    '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)
parser_stop.add_argument(
    'list', help='List of agents to stop, leave empty to stop all agents', metavar='list', type=str, nargs='*',choices=container_names)

parser_update.add_argument(
    '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)
parser_update.add_argument(
    '--reset', help='Triggers new login before update', action='store_true')
parser_update.add_argument("-d", "--detach", help="Runs update in detached mode", action="store_true")

"""Adding an add_argument to parser_update"""
parser_update.add_argument(
    '--image', '-i', dest ='image',help='List of docker images to update', metavar='flag', type=str, nargs='*')  
parser_update.add_argument(
    'list', help='List of agents to update, leave empty to start all agents', metavar='list', type=str, nargs='*',choices=container_names)


parser_orbitty.add_argument(
    '--path', help='Cognicept configuration directory (default: `'
    + DEFAULT_PATH + '`)', default=DEFAULT_PATH)

parser_record.add_argument(
    '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)
parser_record.add_argument(
    '--start',
    help='Start rosbag recording session. Provide topics to record separated by spaces with their `/` prefix. e.g. cognicept record --start /rosout /odom',
    nargs='+')
parser_record.add_argument(
    '--all', help='Start rosbag recording session to record ALL topics.', action='store_true')
parser_record.add_argument(
    '--stop',
    help='Stop rosbag recording session. Set `STOP` to `autopush` to automatically push latest bag to the Cognicept cloud after stopping',
    type=str, const='nopush', nargs='?')
parser_record.add_argument(
    '--pause', help='Pause rosbag recording session', action='store_true')
parser_record.add_argument(
    '--resume', help='Resume rosbag recording session', action='store_true')
parser_record.add_argument(
    '--status', help='Get current recording session status', action='store_true')


parser_push.add_argument(
    '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)
parser_push.add_argument(
    '--bag',
    help='Pushes rosbag to the Cognicept cloud. By default it pushes the latest bag file recording. Set `BAG` to appropriate bag name in the `PATH` folder to upload a rosbag by name',
    type=str, const='latest', nargs='?')


parser_init.add_argument(
    '--path', default=DEFAULT_PATH)
parser_init.add_argument(
    '--robot_id', help='Input the robot ID', required=True)
parser_init.add_argument(
    '--org_id', help='Input the organisation ID', required=True)

argcomplete.autocomplete(parser)

from cogniceptshell.agent_life_cycle import AgentLifeCycle
from cogniceptshell.rosbag_record import RosbagRecord
from cogniceptshell.pusher import Pusher

parser_config.set_defaults(func=local_cfg.configure)
parser_keyrotate.set_defaults(func=local_cfg.cognicept_key_rotate)
agent_lifetime = AgentLifeCycle()
parser_version.set_defaults(func=agent_lifetime.display_version)
parser_status.set_defaults(func=agent_lifetime.status)
parser_lastevent.set_defaults(func=agent_lifetime.get_last_event)
parser_restart.set_defaults(func=agent_lifetime.restart)
parser_start.set_defaults(func=agent_lifetime.start)
parser_stop.set_defaults(func=agent_lifetime.stop)
parser_update.set_defaults(func=agent_lifetime.update)
parser_orbitty.set_defaults(func=agent_lifetime.run_orbitty)
record_session = RosbagRecord()
parser_record.set_defaults(func=record_session.record)
pusher_instance = Pusher()
parser_push.set_defaults(func=pusher_instance.push)
parser_init.set_defaults(func=local_cfg.init_config)

def main():

    # Parse the arguments
    args = parser.parse_args()

    if ("path" not in args):
        parser.print_help()
    else:
        local_cfg.load_config(args.path)
        agent_lifetime.configure_containers(local_cfg)
        args.config = local_cfg

        if (hasattr(args, 'func')):
            args.func(args)


if __name__ == "__main__":
    main()



# Copyright 2020 Cognicept Systems
# Author: Swarooph Seshadri (swarooph@cognicept.systems)
# --> common utilities for the cognicept shell goes here.
import docker

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def generate_progress_bar(iteration, total, decimals=1, length=100, fill='█'):
    """
    Call in a loop to create terminal progress bar
        Parameters:
            iteration   - Required  : current iteration (Int)
            total       - Required  : total iterations (Int)
            decimals    - Optional  : positive number of decimals in percent complete (Int)
            length      - Optional  : character length of bar (Int)
            fill        - Optional  : bar fill character (Str)
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 *
                                                     (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
    output = "|" + bar + "| " + percent + "%"
    return output

class DockerPermissionError():
    identifier_string = "Error while fetching server API version: ('Connection aborted.', PermissionError(13, 'Permission denied'))"
    suggestion = "Error while fetching server API version, Your current user might not have permissions to access the resources. If docker is newly installed try:\n " \
                 "          sudo usermod -aG docker $USER\n" \
                 "Followed by rebooting your system."



def permission_safe_docker_call(func, *args, **kwargs):
    """
    Run a function while catching for docker permissions related exceptions.

            Parameters:
                    func: function to be executed
                    args: list of positional arguments for executing func
                    kwargs: Mapping of keyword arguments for executing func
            Returns:
                    result: None if failed, else returns result for function call
    """
    try:
        result = func(*args,**kwargs)
    except docker.errors.DockerException as docker_exception:
        if str(docker_exception) == DockerPermissionError.identifier_string:
            print(DockerPermissionError.suggestion)
            return None
        else:
            raise docker_exception
    return result



import docker

class MockDockerClient:
    """
    Class to mock docker client
    """

    def __init__(self):
        self.containers = MockDockerContainers()
        self.images = MockDockerImages()


class MockDockerContainers:
    """
    Class to mock docker.containers API
    """

    def __init__(self):
        self._running_containers = []
        self._stopped_containers = []

    def run(self, docker_image, command, **options):
        """
        Mock run container command
        """
        if(not "name" in options):
            raise docker.errors.ContainerError("Container name is empty")
        container_name = options["name"]
        print("called run " + container_name)
        if(container_name in self._running_containers):
            raise docker.errors.ContainerError("Container already exists")
        else:
            self._running_containers.append(container_name)

    def _stop(self, container_name):
        """
        Helper stop container that is passed as lambda in returning docker object
        """
        self._running_containers.remove(container_name)
        self._stopped_containers.append(container_name)

    def _remove(self, container_name):
        """
        Helper remove function that is passed as lambda in returning docker object
        """
        self._stopped_containers.remove(container_name)

    def get(self, container_name):
        """
        Mock get container api
        """
        if(container_name in self._running_containers):
            def stop_function(timeout): return self._stop(container_name)
            def remove_function(): return self._remove(container_name)
            return MockContainer("running", stop_function, remove_function)
        else:
            raise docker.errors.NotFound("Container not found")

class MockContainer:
    """
    Class to mock docker.containers API
    """

    def __init__(self, status, stop_function, remove_function):
        self.status = status
        self.stop = stop_function
        self.remove = remove_function


class MockDockerImages:
    """
    Class to mock docker.images API
    """

    def __init__(self):
        self._images = []

    def get(self, image_name):
        """
        Mock get image api
        """
        if(image_name in self._images):
            return image_name
        else:
            raise docker.errors.ImageNotFound("Image not found")



import mock
import pytest
import requests
from mock import patch
import os
from cogniceptshell.common import bcolors

from cogniceptshell.configuration import Configuration


class SuccessAwsCredentials(object):
    def __init__(self):
        self.status_code = 200

    def json(self):
        return {"AccessKeyId": os.getenv('AWS_ACCESS_KEY_ID', ""), "SecretAccessKey": os.getenv('AWS_SECRET_ACCESS_KEY', ""), "SessionToken": ""}


class NotFoundMockResponse(object):
    def __init__(self):
        self.status_code = 404


def mock_aws_endpoint(*args, **kwargs):
    if(args[0] == "https://test.cognicept.systems/api/agent/v1/aws/assume_role"):
        return SuccessAwsCredentials()
    else:
        return NotFoundMockResponse()

def setup_file(tmpdir):
    p = tmpdir.join("runtime.env")
    p.write("COG_AGENT_CONTAINERS=container1;container2\nCOG_AGENT_IMAGES=image1;image2\nCOGNICEPT_API_URI=https://test.cognicept.systems/api/agent/v1/aws/assume_role\nCOGNICEPT_ACCESS_KEY=CORRECT-KEY")


def setup_wrong_uri_file(tmpdir):
    p = tmpdir.join("runtime.env")
    p.write(
        "COGNICEPT_API_URI=https://www.wronguri.blame\nCOGNICEPT_ACCESS_KEY=INCORRECT-KEY")


def setup_correct_uri_file(tmpdir):
    p = tmpdir.join("runtime.env")
    p.write(
        "COGNICEPT_API_URI=https://test.cognicept.systems/api/agent/v1/\nCOGNICEPT_ACCESS_KEY=CORRECT-KEY")


def setup_wrong_uri_file_for_init(tmpdir):
    p = tmpdir.join("runtime.env")

def setup_correct_uri_file_for_init(tmpdir):
    p = tmpdir.join("runtime.env")
    p.write(
        "COGNICEPT_USER_API_URI=https://test.cognicept.systems/api/agent/v1/")

def mock_shutil_copy_fail_no_space(arg1, arg2):
    test_exp = OSError(28, 'No space left on device')
    raise(test_exp)

def mock_shutil_copy_fail_others(arg1, arg2):
    test_exp = OSError(1, 'Operation not permitted')
    raise(test_exp)

def mock_shutil_copy_success(arg1, arg2):
    pass

def mock_file_write_fail(arg1, arg2):
    raise(PermissionError)

def test_yes_input():
    object = Configuration()
    assert (object._interpret_bool_input("Y") == True)

def test_no_input():
    object = Configuration()
    assert (object._interpret_bool_input("n") == False)

def test_other_input():
    object = Configuration()
    assert (object._interpret_bool_input("g") == None)
    assert (object._interpret_bool_input("%") == None)
    assert (object._interpret_bool_input("1") == None)
    assert (object._interpret_bool_input("akjflakjewr4f56f74ew@!!@$@!$") == None)

def test_is_ssh_disabled(tmpdir):
    setup_file(tmpdir)
    object = Configuration()
    object.load_config(str(tmpdir) + "/")
    assert (object.is_ssh_enabled() == False)    
    object.config["COG_ENABLE_SSH_KEY_AUTH"] = False
    assert (object.is_ssh_enabled() == False)

def test_is_ssh_disabled(tmpdir):
    setup_file(tmpdir)
    object = Configuration()
    object.load_config(str(tmpdir) + "/")
    object.config["COG_ENABLE_SSH_KEY_AUTH"] = True
    assert (object.is_ssh_enabled() == True)


def test_incorrect_cognicept_key_fetch_aws_keys(tmpdir, capsys, monkeypatch):
    setup_wrong_uri_file(tmpdir)
    
    monkeypatch.setattr(requests, "get", mock_aws_endpoint)
    object = Configuration()
    object.load_config(str(tmpdir) + "/")
    try:
        result = object.fetch_aws_keys()
        assert(result == False)
    except:
        pytest.fail("Incorrect Cognicept API URI gave exception", pytrace=True)


def test_correct_cognicept_key_fetch_aws_keys(tmpdir, capsys, monkeypatch):
    setup_correct_uri_file(tmpdir)

    monkeypatch.setattr(requests, "get", mock_aws_endpoint)
    object = Configuration()
    object.load_config(str(tmpdir) + "/")
    try:
        result = object.fetch_aws_keys()
        assert(result == SuccessAwsCredentials().json())
    except:
        pytest.fail("Correct Cognicept API URI gave exception", pytrace=True)


def test_wrong_get_cognicept_api_uri_init(tmpdir, capsys):
    setup_wrong_uri_file_for_init(tmpdir)
    object = Configuration()
    object.load_config(str(tmpdir) + "/")
    result = object.get_cognicept_user_api_uri()
    assert(result == 'https://app.cognicept.systems/api/v1/')


def test_correct_get_cognicept_api_uri_init(tmpdir, capsys):
    setup_correct_uri_file_for_init(tmpdir)
    object = Configuration()
    object.load_config(str(tmpdir) + "/")
    result = object.get_cognicept_user_api_uri()
    assert(result == 'https://test.cognicept.systems/api/agent/v1/')


def test_no_runtime_dotenv(tmpdir, capsys):
    object = Configuration()
    object.load_config(str(tmpdir) + "/")
    captured = capsys.readouterr()
    assert str(captured.out) == "Configuration file `" + str(tmpdir) + "/runtime.env` does not exist.\nConfiguration file `" + str(tmpdir) + "/runtime.env` is empty or could not be parsed.\n"
    
def test_with_runtime_dotenv(tmpdir, capsys):
    setup_file(tmpdir)
    object = Configuration()
    result = object.load_config(str(tmpdir) + "/")
    assert result == True

def test_save_config_success(tmpdir, capsys):
    setup_file(tmpdir)
    object = Configuration()
    result = object.load_config(str(tmpdir) + "/")
    object.save_config()
    captured = capsys.readouterr()
    assert str(captured.out) == bcolors.OKBLUE + "Backed up runtime configuration to: " + str(tmpdir) + "/runtime.env.bk" + bcolors.ENDC + "\n"

def test_save_config_fail_backup_no_space(tmpdir, capsys):
    setup_file(tmpdir)
    object = Configuration()
    result = object.load_config(str(tmpdir) + "/")
    with patch('shutil.copyfile', new=mock_shutil_copy_fail_no_space):
        object.save_config()
        captured = capsys.readouterr()
    assert str(captured.out) == bcolors.FAIL + "Could not back up runtime configuration. Aborting saving runtime configuration. No space left on device!" + bcolors.ENDC + "\n"

def test_save_config_fail_backup_other_exceptions(tmpdir, capsys):
    setup_file(tmpdir)
    object = Configuration()
    result = object.load_config(str(tmpdir) + "/")
    with patch('shutil.copyfile', new=mock_shutil_copy_fail_others):
        object.save_config()
        captured = capsys.readouterr()
    assert str(captured.out) == bcolors.FAIL + "Could not back up runtime configuration. Aborting saving runtime configuration: Operation not permitted" + bcolors.ENDC + "\n"

def test_save_config_fail_permissions(tmpdir, capsys):
    setup_file(tmpdir)
    object = Configuration()
    result = object.load_config(str(tmpdir) + "/")
    # Mock shutil to succeed blindly to trigger 
    with patch('shutil.copyfile', new=mock_shutil_copy_success):
        with mock.patch("builtins.open", mock_file_write_fail):
            object.save_config()
            captured = capsys.readouterr()
    assert str(captured.out) == bcolors.OKBLUE + "Backed up runtime configuration to: " + str(tmpdir) + "/runtime.env.bk" + bcolors.ENDC + "\n" + \
                                bcolors.FAIL + "Could not write into `" + str(tmpdir) + "/runtime.env" + "`. Please check write permission or run with `sudo`." + bcolors.ENDC + "\n"
    
def test_get_docker_compose(monkeypatch):
    object = Configuration()
    # Mock config dictionary with COG_COMPOSE_DIR key
    object.config = {"COG_COMPOSE_FILE": "/path/to/docker-compose.yml"}
    
    # Mock yaml file contents
    yaml_data = """
    services:
        camera_top:
            image: 412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/ros:realsense
        ip_cam_front:
            image: 412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/ros:crystal
    """
    
    # Mock the open() function to return yaml file contents
    mock_file = mock.mock_open(read_data=yaml_data)
    monkeypatch.setattr("builtins.open", mock_file)
    # Call the function with the mock config
    result = object.get_docker_compose()
    
    # Check that the expected dictionary was returned
    expected_result = {"camera_top": "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/ros:realsense", "ip_cam_front": "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/ros:crystal"}
    assert result == expected_result

def test_fail_get_docker_compose_(monkeypatch, capsys):
    object = Configuration()
    # Mock config dictionary with COG_COMPOSE_DIR key
    object.config = {"COG_COMPOSE_FILE": "/path/to/docker-compose.yml"}
    
    # Call the function with the input directory but file not found
    result = object.get_docker_compose()
    assert result == {}

    # Mock empty file contents
    yaml_data = """"""
    
    # Mock the open() function to return yaml file contents
    mock_file = mock.mock_open(read_data=yaml_data)
    monkeypatch.setattr("builtins.open", mock_file)
    # Call the function with the mock config with empty data
    result = object.get_docker_compose()
    assert result == {}
    
    # Test wrong try get data from invalid yaml format docker-compose file
    yaml_data = """
    services:
    camera_top:
            image: 412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/ros:realsense
        ip_cam_front:
            image: 412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/ros:crystal
    """
    mock_file1 = mock.mock_open(read_data=yaml_data)
    monkeypatch.setattr("builtins.open", mock_file1)
    result = object.get_docker_compose()
    assert result == {}